package pdf

import (
	"encoding/hex"
	"fmt"
	"strconv"
	"unicode"
)

// isWhitespace returns true if the given byte is a whitespace character.
// PDF defines whitespace as any of the following characters: 0x00, 0x09, 0x0A, 0x0C, 0x0D, 0x20.
// Specification 7.2.3 (Character Set) of the PDF 32000-1:2008 specification.
func isWhitespace(ch rune) bool {
	switch ch {
	case 0x00, 0x09, 0x0A, 0x0C, 0x0D, 0x20:
		return true
	default:
		return false
	}
}

// isDelimiter checks if the given byte is a delimiter character.
// PDF defines delimiters as any of the following characters: '(', ')', '<', '>', '[', ']', '/', '%'.
// Specification 7.2.3 (Character Set) of the PDF 32000-1:2008 specification.
func isDelimiter(ch rune) bool {
	switch ch {
	case '(', ')', '<', '>', '[', ']', '/', '%':
		return true
	default:
		return false
	}
}

// IsEndOfLine checks if the given byte is an end-of-line character.
// PDF defines end-of-line as any of the following characters: 0x0A, 0x0D.
// Specification 7.2.3 (Character Set) of the PDF 32000-1:2008 specification.
func isEndOfLine(ch rune) bool {
	return ch == 0x0A || ch == 0x0D
}

type delimiterType int

const (
	leftParenthesis    delimiterType = iota + 1 // '('
	rightParenthesis                            // ')'
	lessThanSign                                // '<'
	greaterThanSign                             // '>'
	leftSquareBracket                           // '['
	rightSquareBracket                          // ']'
	solidus                                     // '/'
	percentSign                                 // '%'
	unknownDelimiter                            // Unbekannter Delimiter
)

func octalToRune(octal []rune) rune {
	octalString := string(octal)

	value, err := strconv.ParseInt(octalString, 8, 32)
	if err != nil {
		fmt.Println("Konvertierungsfehler:", err)
		return 0
	}

	return rune(value)
}

func hexToString(hexStr string) (string, error) {

	if len(hexStr)%2 != 0 {
		hexStr += "0"
	}

	bytes, err := hex.DecodeString(hexStr)
	if err != nil {
		return "", err
	}

	return string(bytes), nil
}

func isLeftParenthesis(ch rune) bool {
	return ch == '('
}

func isRightParenthesis(ch rune) bool {
	return ch == ')'
}

func isLessThanSign(ch rune) bool {
	return ch == '<'
}

func isGreaterThanSign(ch rune) bool {
	return ch == '>'
}

func isLeftSquareBracket(ch rune) bool {
	return ch == '['
}

func isRightSquareBracket(ch rune) bool {
	return ch == ']'
}

func isSolidus(ch rune) bool {
	return ch == '/'
}

func isPercentSign(ch rune) bool {
	return ch == '%'
}

func isReverseSolidus(ch rune) bool {
	return ch == '\\'
}

func isNumeric(ch rune) bool {
	return unicode.IsDigit(ch)
}

func getDelimiterType(ch rune) delimiterType {
	switch ch {
	case '(':
		return leftParenthesis
	case ')':
		return rightParenthesis
	case '<':
		return lessThanSign
	case '>':
		return greaterThanSign
	case '[':
		return leftSquareBracket
	case ']':
		return rightSquareBracket
	case '/':
		return solidus
	case '%':
		return percentSign
	default:
		return unknownDelimiter
	}
}
