package pdf

type stringToken struct {
	tokenInternal
	value string
}

type readStringObjectState struct {
	token *stringToken
}

const (
	ModeUnknown = iota
	ModeLiteral = iota
	ModeHex
)

func (s readStringObjectState) process(t *Tokenizer) {

	var c rune
	var err error

	var value string

	mode := ModeUnknown

	parenthesesCount := 0

	for {

		c, err = t.consumeNextRune()
		if err != nil {
			t.raiseError(err)
			return
		}

		if isEndOfFile(c) {
			t.raiseError(ErrUnexpectedEOF)
			return
		}

		if mode == ModeUnknown {
			switch {
			case isLeftParenthesis(c):
				mode = ModeLiteral
				continue
			case isLessThanSign(c):
				mode = ModeHex
				continue
			}
		}

		if mode == ModeLiteral {
			if parenthesesCount == 0 && isRightParenthesis(c) {
				break
			}

			// Within a literal string, the REVERSE SOLIDUS is used as an escape character. The character
			//immediately following the REVERSE SOLIDUS determines its precise interpretation as shown in "Table
			//3 — Escape sequences in literal strings". If the character following the REVERSE SOLIDUS is not one of
			//those shown in "Table 3 — Escape sequences in literal strings", the REVERSE SOLIDUS shall be
			//ignored.
			if isReverseSolidus(c) {
				c, err = t.consumeNextRune()

				if err != nil {
					t.raiseError(err)
					return
				}
				if isEndOfFile(c) {
					t.raiseError(ErrUnexpectedEOF)
					return
				}

				// A PDF writer may split a literal string across multiple lines. The REVERSE SOLIDUS (5Ch) (backslash
				//character) at the end of a line shall be used to indicate that the string continues on the following line. A
				//PDF processor shall disregard the REVERSE SOLIDUS and the end-of-line marker following it when
				//reading the string; the resulting string value shall be identical to that which would be read if the string
				//were not split.
				if isEndOfLine(c) {
					c, err = t.consumeNextRune()
					if err != nil {
						t.raiseError(err)
						return
					}
					if isEndOfFile(c) {
						t.raiseError(ErrUnexpectedEOF)
						return
					}
					if !isEndOfLine(c) {
						err = t.unread()
						if err != nil {
							t.raiseError(err)
							return
						}
					}
					continue
				}

				switch {
				case c == 'n':
					value += "\n"
					continue
				case c == 'r':
					value += "\r"
					continue
				case c == 't':
					value += "\t"
					continue
				case c == 'b':
					value += "\b"
					continue
				case c == 'f':
					value += "\f"
					continue
				case c == '(':
					value += "("
					continue
				case c == ')':
					value += ")"
					continue
				case c == '\\':
					value += "\\"
					continue
				}

				if isNumeric(c) {
					numConter := 0
					var digit []rune
					for {
						digit = append(digit, c)
						numConter++
						if numConter == 3 {
							break
						}

						c, err = t.consumeNextRune()
						if err != nil {
							t.raiseError(err)
							return
						}
						if isEndOfFile(c) {
							t.raiseError(ErrUnexpectedEOF)
							return
						}

						if !isNumeric(c) {
							err := t.unread()
							if err != nil {
								t.raiseError(err)
								return
							}

							break
						}
					}

					// Convert the octal number to a character
					// The octal number is made up of one, two, or three octal digits (0-7).
					value += string(octalToRune(digit))
					continue

				}
			}

			if isLeftParenthesis(c) {
				parenthesesCount++
			} else if isRightParenthesis(c) {
				parenthesesCount--
				if parenthesesCount < 0 {
					t.raiseError(ErrInvalidParenthesesCount)
					return
				}
			}

			if isEndOfLine(c) {
				c, err = t.consumeNextRune()
				if err != nil {
					t.raiseError(err)
					return
				}

				if isEndOfFile(c) {
					t.raiseError(ErrUnexpectedEOF)
					return
				}

				if !isEndOfLine(c) {
					err := t.unread()
					if err != nil {
						t.raiseError(err)
						return
					}
				}
				c = '\n'
			}

			value += string(c)
		} else if mode == ModeHex {
			if parenthesesCount == 0 && isGreaterThanSign(c) {
				break
			}

			value += string(c)

		}

	}

	if mode == ModeHex {
		value, err = hexToString(value)
		if err != nil {
			t.raiseError(err)
			return
		}
	}

	s.token.value = value
	t.emitToken(s.token)

	return
}
