package pdf

import (
	"bufio"
	"fmt"
	"io"
	"reflect"
	"sync"
)

func NewTokenizer(r io.Reader) *Tokenizer {
	return &Tokenizer{
		reader:     bufio.NewReader(r),
		tokenStack: make([]tokenInterface, 0),
		parseError: make([]error, 0),
		//state:      initialState{},
		//tmpBuf: "",
		tmpBuf:   []rune{},
		lastRune: []rune{},
		mu:       sync.Mutex{},
	}
}

//type StateType int

//const (
//	InitialStateType StateType = iota + 1
//	CommentStateType
//	WhitespaceStateType
//	RegularStateType
//	NumberStateType
//	NameStateType
//	LiteralStringStateType
//	HexStringStateType
//	KeywordStateType
//)

type Tokenizer struct {
	reader     *bufio.Reader
	tokenStack []tokenInterface
	parseError []error

	state state

	byteOffset int

	lastRune     []rune
	lastRuneSize int
	tmpBuf       []rune

	mu sync.Mutex
}

//func (t *Tokenizer) NextToken() (*tokenInterface, error) {
//	line, err := t.reader.ReadString('\n')
//	if err != nil {
//		return nil, err
//	}
//	line = strings.TrimSpace(line)
//	if strings.HasPrefix(line, "%") {
//		return &tokenInterface{Type: Comment, Value: line}, nil
//	}
//	// Füge hier weitere Regeln hinzu, um andere Typen von tokenInterface zu identifizieren
//	return &tokenInterface{Type: Unknown, Value: line}, nil
//}

func (t *Tokenizer) raiseError(err error) {
	t.mu.Lock()
	defer t.mu.Unlock()

	t.parseError = append(t.parseError, err)
}

func (t *Tokenizer) Errors() []error {
	t.mu.Lock()
	defer t.mu.Unlock()

	return t.parseError
}

func (t *Tokenizer) HasError() bool {
	t.mu.Lock()
	defer t.mu.Unlock()

	return len(t.parseError) > 0
}

//func (t *Tokenizer) GetTokenStack() []tokenInterface {
//	return t.tokenStack
//}

func (t *Tokenizer) PushToken(o tokenInterface) {
	t.mu.Lock()
	defer t.mu.Unlock()

	t.tokenStack = append(t.tokenStack, o)
}

func (t *Tokenizer) popToken() *tokenInterface {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.tokenStack) == 0 {
		return nil
	}
	token := t.tokenStack[len(t.tokenStack)-1]
	t.tokenStack = t.tokenStack[:len(t.tokenStack)-1]
	return &token
}

func (t *Tokenizer) peekToken() *tokenInterface {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.tokenStack) == 0 {
		return nil
	}
	return &t.tokenStack[len(t.tokenStack)-1]
}

func (t *Tokenizer) emitToken(o tokenInterface) {
	if o == nil || (reflect.ValueOf(o).IsNil()) {
		t.raiseError(ErrorTokenNil)
		return
	}

	if o.isEmitted() {
		t.raiseError(ErrTokenAlreadyEmitted)
		return
	}

	o.emit()
	o.setByteOffset(int64(t.byteOffset))
	t.PushToken(o)
}

//func (t *Tokenizer) tokenCount() int {
//	return len(t.tokenStack)
//}

//func (t *Tokenizer) resetTempBuffer() {
//	t.tmpBuf = []rune{}
//}
//
//func (t *Tokenizer) appendToTempBuffer(c rune) {
//	t.tmpBuf = append(t.tmpBuf, c)
//	t.byteOffset += len([]byte(string(c)))
//}

//func (t *Tokenizer) setTempBuffer(r []rune) {
//	t.tmpBuf = r
//	t.byteOffset -= len(r)
//}

//func (t *Tokenizer) tempBuffer() string {
//	return string(t.tmpBuf)
//}

//func (t *Tokenizer) getTempBuffer() []rune {
//	return t.tmpBuf
//}

func (t *Tokenizer) peekNextRune() (rune, error) {
	r, err := t.consumeNextRune()
	if err != nil {
		return 0, err
	}
	err = t.unread()
	if err != nil {
		return 0, err
	}
	return r, nil
}

func (t *Tokenizer) peekNextRunes(count int) ([]rune, error) {
	runes := make([]rune, 0)
	for i := 0; i < count; i++ {
		r, err := t.consumeNextRune()
		if err != nil {
			return nil, err
		}
		runes = append(runes, r)
	}
	err := t.unreadCount(count)
	if err != nil {
		return nil, err
	}
	return runes, nil
}

func (t *Tokenizer) consumeNextRune() (rune, error) {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.tmpBuf) > 0 {
		c := t.tmpBuf[0]
		t.tmpBuf = t.tmpBuf[1:]

		t.byteOffset += len([]byte(string(c)))
		t.lastRune = append(t.lastRune, c)

		fmt.Println("consumeNextRune (tmp): ", c, string(c), t.byteOffset)

		return c, nil
	}

	c, s, err := t.reader.ReadRune()
	if err != nil {
		return 0, err
	}

	t.byteOffset += len([]byte(string(c)))
	t.lastRune = append(t.lastRune, c)
	t.lastRuneSize = s

	fmt.Println("consumeNextRune: ", c, string(c), t.byteOffset)

	return c, nil
}

func (t *Tokenizer) unread() error {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.lastRune) == 0 {
		return ErrNoRuneToUnread
	}

	u := t.lastRune[len(t.lastRune)-1]

	r := t.lastRune[:len(t.lastRune)-1]
	t.byteOffset -= len([]byte(string(u)))
	t.lastRune = r

	t.tmpBuf = append([]rune{u}, t.tmpBuf...)

	fmt.Println("unread: ", u, string(u), t.byteOffset)

	return nil
}

func (t *Tokenizer) unreadCount(count int) error {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.lastRune) < count {
		return ErrNoRuneToUnread
	}

	for i := 0; i < count; i++ {
		u := t.lastRune[len(t.lastRune)-1]
		t.tmpBuf = append([]rune{u}, t.tmpBuf...)
		t.lastRune = t.lastRune[:len(t.lastRune)-1]
		t.byteOffset -= len([]byte(string(u)))
	}

	fmt.Println("unreadCount: ", count, t.byteOffset)

	return nil
}

//func (t *Tokenizer) reconsumeCharacter(count int) {
//	for i := 0; i < count; i++ {
//		_ = t.reader.UnreadRune()
//	}
//}

func (t *Tokenizer) switchState(state state) {
	t.mu.Lock()
	t.state = state
	t.mu.Unlock()

	state.process(t)
}

//	func (t *Tokenizer) currentState() state {
//		return t.state
//	}
func (t *Tokenizer) getLastToken() tokenInterface {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.tokenStack) == 0 {
		return nil
	}
	return t.tokenStack[len(t.tokenStack)-1]
}

func (t *Tokenizer) penultimateToken() tokenInterface {
	t.mu.Lock()
	defer t.mu.Unlock()

	if len(t.tokenStack) < 2 {
		return nil
	}
	return t.tokenStack[len(t.tokenStack)-2]
}
