package pdf

import (
	"errors"
	"io"
	"os"
	"testing"
)

var ErrTokenizerTest = errors.New("invalid")

func createTestTokenizer(t *testing.T) *Tokenizer {
	file, err := os.Open("assets/Simple-PDF-2.0-file.pdf")
	if err != nil {
		t.Errorf("Expected to open the file, got %v", err)
		os.Exit(1)
	}

	tok := NewTokenizer(file)
	if tok == nil {
		t.Errorf("Expected NewTokenizer to return a non-nil value")
	}

	return tok
}

// Examples https://github.com/pdf-association/pdf20examples
func TestNewTokenizer(t *testing.T) {

	file, err := os.Open("assets/Simple-PDF-2.0-file.pdf")
	if err != nil {
		t.Errorf("Expected to open the file, got %v", err)
		os.Exit(1)
	}

	tok := NewTokenizer(file)
	if tok == nil {
		t.Errorf("Expected NewTokenizer to return a non-nil value")
	}
}

func TestTokenizerRaiseError(t *testing.T) {

	tok := createTestTokenizer(t)

	if tok.HasError() == true {
		t.Errorf("Exspected true, got false")
	}

	tok.raiseError(ErrTokenizerTest)
	if tok.HasError() == false {
		t.Errorf("Exspected false, got true")
	}

}

type testReaderForTestTokenizer struct {
	data []byte // Testdaten, die gelesen werden sollen
	pos  int    // Aktuelle Position im Daten-Array
}

func (m *testReaderForTestTokenizer) Read(p []byte) (n int, err error) {
	if m.pos >= len(m.data) {
		return 0, io.EOF // Wenn alle Daten gelesen wurden, EOF zurückgeben
	}

	n = copy(p, m.data[m.pos:])
	m.pos += n
	return n, nil
}

func createTestTokenizerWithData(data string) *Tokenizer {
	tr := &testReaderForTestTokenizer{data: []byte(data)}
	return NewTokenizer(tr)
}

func TestTokenizerConsumeNextInputRune(t *testing.T) {
	tests := []struct {
		name     string
		input    string
		expected rune
		hasError error
	}{
		{"Empty input", "", 0, io.EOF},
		{"Single character", "a", 'a', nil},
		{"Multiple characters", "abc", 'a', nil},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tok := createTestTokenizerWithData(tt.input)
			r, err := tok.consumeNextRune()

			if tt.hasError != nil {
				if !errors.Is(err, tt.hasError) {
					t.Errorf("expected error %v, got %v", tt.hasError, err)
				}
			} else {
				if r != tt.expected {
					t.Errorf("expected %v, got %v", tt.expected, r)
				}
			}
		})
	}
}

func TestTokenizerUnread(t *testing.T) {
	tests := []struct {
		name     string
		input    string
		expected rune
	}{
		{"Single character unread", "a", 'a'},
		{"Multiple characters unread", "abc", 'c'},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tok := createTestTokenizerWithData(tt.input)

			// Lese alle Zeichen bis zum Ende
			for range tt.input {
				_, _ = tok.consumeNextRune()
			}

			// Rückgängigmachen des letzten Zeichens
			err := tok.unread()
			if err != nil {
				t.Errorf("expected no error, got %v", err)
			}

			// Das letzte Zeichen sollte das erwartete sein
			r, err := tok.consumeNextRune()
			if err != nil {
				t.Errorf("expected no error, got %v", err)
			}

			if r != tt.expected {
				t.Errorf("expected rune %v, got rune %v", tt.expected, r)
			}
		})
	}
}

func TestTokenizerUnreadCount(t *testing.T) {
	tests := []struct {
		name          string
		input         string
		unreadCount   int
		expectedRune  rune
		expectedError error
	}{
		{"Unread 1 character", "abc", 1, 'c', nil},
		{"Unread 2 characters", "abc", 2, 'b', nil},
		{"Unread more characters than available", "a", 2, 0, ErrNoRuneToUnread},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tok := createTestTokenizerWithData(tt.input)

			// Lese alle Zeichen bis zum Ende
			for range tt.input {
				_, _ = tok.consumeNextRune()
			}

			// Rückgängigmachen von `unreadCount` Zeichen
			err := tok.unreadCount(tt.unreadCount)
			if tt.expectedError != nil {
				if !errors.Is(err, tt.expectedError) {
					t.Errorf("expected error %v, got %v", tt.expectedError, err)
				}
				return // Kein weiterer Test nötig, da wir einen Fehler erwarten
			}

			if err != nil {
				t.Errorf("expected no error, got %v", err)
			}

			// Prüfen, ob das nächste Zeichen korrekt gelesen wird
			r, err := tok.consumeNextRune()
			if err != nil {
				t.Errorf("expected no error, got %v", err)
			}

			if r != tt.expectedRune {
				t.Errorf("expected rune %v, got rune %v", tt.expectedRune, r)
			}
		})
	}
}

func TestTokenizerUnreadWithoutReading(t *testing.T) {
	tok := createTestTokenizerWithData("a")

	// Versuche, ein Zeichen rückgängig zu machen, ohne vorher zu lesen
	err := tok.unread()
	if !errors.Is(err, ErrNoRuneToUnread) {
		t.Errorf("expected error %v, got %v", ErrNoRuneToUnread, err)
	}
}

func TestTokenizerUnreadCountWithoutEnoughRunes(t *testing.T) {
	tok := createTestTokenizerWithData("ab")

	// Lese ein Zeichen
	_, _ = tok.consumeNextRune()

	// Versuche, zwei Zeichen rückgängig zu machen, obwohl nur eins gelesen wurde
	err := tok.unreadCount(2)
	if !errors.Is(err, ErrNoRuneToUnread) {
		t.Errorf("expected error %v, got %v", ErrNoRuneToUnread, err)
	}
}

func TestTokenizerConsumeAfterUnread(t *testing.T) {
	tok := createTestTokenizerWithData("abc")

	// Lese zwei Zeichen: 'a', 'b'
	r1, _ := tok.consumeNextRune()
	r2, _ := tok.consumeNextRune()

	// Unread das letzte Zeichen ('b')
	_ = tok.unread()

	// Das nächste Zeichen sollte wieder 'b' sein
	r3, _ := tok.consumeNextRune()

	if r3 != r2 {
		t.Errorf("expected rune %v after unread, but got %v", r2, r3)
	}

	// Das erste Zeichen sollte weiterhin 'a' sein
	if r1 != 'a' {
		t.Errorf("first read rune should still be %v, but got %v", 'a', r1)
	}
}

type mockToken struct {
	tokenInternal
}

type mockState struct {
	wasProcessed bool
}

func (m *mockState) process(t *Tokenizer) {
	m.wasProcessed = true
}

func TestTokenizerSwitchState(t *testing.T) {
	// Mock-State für den Test
	mockState := &mockState{}

	tok := createTestTokenizerWithData("")

	// Wechsle zum neuen Mock-Zustand
	tok.switchState(mockState)

	if tok.state != mockState {
		t.Errorf("expected state to be %v, got %v", mockState, tok.state)
	}

	if !mockState.wasProcessed {
		t.Errorf("expected state's process function to be called, but it was not")
	}
}

func TestTokenizerGetLastToken(t *testing.T) {
	tok := createTestTokenizerWithData("")

	// Teste für den Fall, dass der token-Stack leer ist
	token := tok.getLastToken()
	if token != nil {
		t.Errorf("expected nil when no tokens are available, got %v", token)
	}

	// Teste, wenn Tokens auf dem Stack vorhanden sind
	mockToken := &mockToken{}
	tok.tokenStack = append(tok.tokenStack, mockToken)

	lastToken := tok.getLastToken()
	if lastToken != mockToken {
		t.Errorf("expected last token to be %v, got %v", mockToken, lastToken)
	}
}

func TestPeekNextRune(t *testing.T) {
	tests := []struct {
		name     string
		input    string
		expected rune
	}{
		{"Single character", "a", 'a'},
		{"Multiple characters", "abc", 'a'},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tok := createTestTokenizerWithData(tt.input)

			r1, _ := tok.peekNextRune()
			r2, _ := tok.peekNextRune()
			r3, _ := tok.peekNextRune()
			r4, _ := tok.peekNextRune()

			if r1 != tt.expected && r2 != tt.expected && r3 != tt.expected && r4 != tt.expected {
				t.Errorf("expected rune %v, got %v %v %v %v", tt.expected, r1, r2, r3, r4)
			}
		})
	}
}

func TestPeekNextRunes(t *testing.T) {
	tests := []struct {
		name          string
		input         string
		count         int
		expected      string
		expectedError error
	}{
		{"Single character", "a", 1, "a", nil},
		{"Multiple characters", "abc", 2, "ab", nil},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tok := createTestTokenizerWithData(tt.input)

			r, e := tok.peekNextRunes(tt.count)
			if tt.expectedError != nil {
				if !errors.Is(e, tt.expectedError) {
					t.Errorf("expected error %v, got %v", tt.expectedError, e)
				}
				return
			}

			if string(r) != tt.expected {
				t.Errorf("expected %v, got %v", tt.expected, string(r))
			}

		})
	}
}
